/*
   This file shows how to use "pointers to void",
   along with "function pointers", to implement a
   "generic" list in C.

   Every list can hold any type of data, so if you
   want to, you could mix up ints, doubles and
   strings in a single list. But if you want to be
   able to print out a list, then the list must be
   homogeneous because there has to be a single
   "toString" function that knows how to print the
   data items in the list.
*/
#include <stdio.h>
#include <stdlib.h> // exit()

struct Node {
   void *data;
   struct Node *tail;
};


int empty(struct Node *list) {
   return NULL == list;
}


void* head(struct Node *list) {
   if (NULL == list)
      exit(-1);  // C does not have exceptions

   return list->data;
}


struct Node* tail(struct Node *list) {
   if (NULL == list)
      exit(-1);  // C does not have exceptions

   return list->tail;
}


struct Node* cons(void *data, struct Node *list) {
   struct Node *newNode = (struct Node*)malloc(sizeof(struct Node));
   newNode->data = data;
   newNode->tail = list;
   return newNode;
}


int length(struct Node *list) {
   int counter = 0;
   while (NULL != list) {
      counter++;
      list = list->tail; // (*list).tail;
   }
   return counter;
}


int lengthRec(struct Node *list) {
   if (NULL == list) // empty list
      return 0;
   else              // non-empty list
      return 1 + lengthRec(list->tail); // recursion
}


// This is a "generic" function that takes a pointer to a function
// that knows how to compare two pieces of data from the list.
int lookUp(void *data, struct Node *list, int equals(void*, void*)) {
   if (NULL == data)
      exit(-1);  // C does not have exceptions

   if (NULL == list)                  // empty list
      return 0;
   else if (equals(data, list->data)) // non-empty list
      return 1;
   else
      return lookUp(data, list->tail, equals); // recursion
}


// This is a "generic" function that takes a pointer to a function
// that knows how to compare two pieces of data from the lists.
int equals(struct Node *list1, struct Node *list2, int equal(void*, void*)) {
   if (list1 == list2) return 1;
   if (NULL == list1 || NULL == list2) return 0;
   while (NULL != list1 && NULL != list2) {
      if ( ! equal(list1->data, list2->data) )
         return 0;
      list1 = list1->tail;
      list2 = list2->tail;
   }
   if (NULL != list1 || NULL != list2) return 0;
   return 1;
}


// This is a "generic" function that takes a pointer to a function
// that knows how to compare two pieces of data from the lists.
int equalsRec(struct Node *list1, struct Node *list2, int equal(void*, void*)) {
   if (list1 == list2)
      return 1;
   else if (NULL == list1 || NULL == list2) // one list is empty but the other isn't
      return 0;
   else if ( ! equal(list1->data, list2->data) )
      return 0;
   else
      return equalsRec(list1->tail, list2->tail, equal); // recursion
}


// This is a "generic" function that takes a pointer to a
// function that knows how to print the data in the list.
void printList(struct Node *list, void toString(void*)) {
   printf("[");
   if (NULL != list) {
      toString(list->data);
      list = list->tail;
   }
   while (NULL != list) {
      printf(", ");
      toString(list->data);
      list = list->tail;
   }
   printf("]\n");
}


// Here are some example functions that can be used by
// the generic printList() function.
void toStringInt(void *x) // x is really a int*
{
   printf( "%d", *((int*)x) );
}
void toStringDouble(void *x) // x is really a double*
{
   printf( "%f", *((double*)x) );
}
void toStringString(void *x) // x is really a char*
{
   printf( "%s", (char*)x );
}


// Here are some example functions that can be used by
// the generic lookUp() function or the generic equals().
int equalInt(void *x, void *y) // x and y are really int*
{
   return *((int*)x) == *((int*)y);
}
int equalDouble(void *x, void *y) // x and y are really double*
{
   return *((double*)x) == *((double*)y);
}
int equalString(void *x, void *y) // x and y are really char*
{
   if (x == y) return 1;

   if (NULL == x || NULL == y) return 0;

   char *p1 = (char*)x, *p2 = (char*)y;

   while ('\0' != *p1 && '\0' != *p2) {
      if (*p1 != *p2) return 0;
      p1 += 1;
      p2 += 1;
   }

   if ('\0' != *p1 || '\0' != *p2) return 0;

   return 1;
}
