/*
   This file implements a generic list with a polymorphic
   toString() function that can print out (almost) any
   list and a polymorphic equals() function that compares
   two lists.

   This file shows how to use "pointers to void",
   along with "function pointers", to implement a
   "generic", "polymorphic" list in C.

   Every list can hold any type of data, so if you
   want to, you can mix up ints, doubles and strings
   in a single list. This version of the list is able
   to print out any list, as long as each "object" in
   the list knows how to print itself. Each object in
   the list must have a pointer to a "toString" function
   that knows how to print the data in that object.

   This example gives you a very rough idea of how
   polymorphic methods are implemented by an
   Object-Oriented Programming language (they use
   a lot of function pointers).
*/
#include <stdio.h>
#include <stdlib.h> // exit()
#include "object.h"

struct Node {
   struct Object *object; // notice that the data is now in another struct
   struct Node *tail;
};


int empty(struct Node *list) {
   return NULL == list;
}


struct Object* head(struct Node *list) {
   if (NULL == list)
      exit(-1);  // C does not have exceptions

   return list->object;
}


struct Node* tail(struct Node *list) {
   if (NULL == list)
      exit(-1);  // C does not have exceptions

   return list->tail;
}


struct Node* cons(struct Object *object, struct Node *list) {
   struct Node *newNode = (struct Node*)malloc(sizeof(struct Node));
   newNode->object = object;
   newNode->tail = list;
   return newNode;
}


int length(struct Node *list)
{
   int counter = 0;
   while (NULL != list)
   {
      counter++;
      list = list->tail; // (*list).tail;
   }
   return counter;
}


int lengthRec(struct Node *list) {
   if (NULL == list) // empty list
      return 0;
   else              // non-empty list
      return 1 + lengthRec(list->tail);
}


// This is a "generic" function that determines if a value is in a list.
// Each "object" in a list is responsible for knowing how to compare itself.
int lookUp(void *data, struct Node *list) {
   if (NULL == data)
      exit(-1);  // C does not have exceptions

   if (NULL == list)    // empty list
      return 0;
   else if ((list->object->equals)(list->object->data, data)) // non-empty list
      return 1;
   else
      return lookUp(data, list->tail); // recursion
}


// This is a "generic" function that determines if two lists are equal.
// Each "object" in a list is responsible for knowing how to compare itself.
int equals(struct Node *list1, struct Node *list2) {
   if (list1 == list2) return 1;
   if (NULL == list1 || NULL == list2) return 0;
   while (NULL != list1 && NULL != list2) {
      if ( ! (list1->object->equals)(list1->object->data, list2->object->data) )
         return 0;
      list1 = list1->tail;
      list2 = list2->tail;
   }
   if (NULL != list1 || NULL != list2) return 0;
   return 1;
}


// This is a "generic" function that recursively determines if two lists are equal.
// Each "object" in a list is responsible for knowing how to compare itself.
int equalsRec(struct Node *list1, struct Node *list2) {
   if (list1 == list2)
      return 1;
   else if (NULL == list1 || NULL == list2) // one list is empty but the other isn't
      return 0;
   else if ( ! (list1->object->equals)(list1->object->data, list2->object->data) )
      return 0;
   else
      return equalsRec(list1->tail, list2->tail); // recursion
}


// This is a "generic" function that prints the data in a list. Each
// "object" in the list is responsible for knowing how to print itself.
void printList(struct Node *list) {
   printf("[");
   if (NULL != list) {
      (list->object->toString)(list->object->data); // easy to read, right!
      list = list->tail;
   }
   while (NULL != list) {
      printf(", ");
      (list->object->toString)(list->object->data); // easy to read, right!
      list = list->tail;
   }
   printf("]\n");
}
