/*
   This file implements a generic list with polymorphic functions.

   This file shows how to use "pointers to void", along with
   "function pointers", to implement a "generic", "polymorphic"
   list in C.

   Every list can hold any type of data, so if you
   want to, you can mix up ints, doubles and strings
   in a single list. This version of the list is able
   to print out any list, as long as each "object" in
   the list knows how to print itself. Each object in
   the list must have a pointer to a "toString" function
   that knows how to print the data in that object.

   This example gives you a very rough idea of how
   polymorphic methods are implemented by an
   Object-Oriented Programming language (they use
   a lot of function pointers).
*/
#include <stdio.h>
#include <stdlib.h> // exit()
#include "node.h"
#include "List.h"


List* newList() {  // construct an empty List
   List *list = (List*)malloc(sizeof(List));
   list->head = NULL;
   return list;
}


void freeList(List *list) {
   // deallocate the list
   struct Node *temp = (struct Node*)(list->head);
   while (NULL != temp) {
      struct Node *temp2 = temp->tail;
      free(temp->object);
      free(temp);
      temp = temp2;
   }
   free(list);
}


int empty(List *list) {
   return NULL == list->head;
}


struct Object* head(List *list) {
   if (NULL == list->head)
      exit(-1);  // C does not have exceptions

   return ((struct Node*)(list->head))->object;
}


List* tail(List *list) {
   if (NULL == list->head)
      exit(-1);  // C does not have exceptions

   List *newList = (List*)malloc(sizeof(List));
   newList->head = ((struct Node*)(list->head))->tail;
   return newList;
}


List* cons(struct Object *object, List *list) {
   struct Node *newNode = (struct Node*)malloc(sizeof(struct Node));
   newNode->object = object;
   newNode->tail = (struct Node*)(list->head);
   List *newList = (List*)malloc(sizeof(List));
   newList->head = newNode;
   return newList;
}


int length(List *list) {
   if (NULL == list->head) // empty list
      return 0;
   else                    // non-empty list
      return 1 + length( tail(list) );
}


// This is a "generic" function that determines if a value is in a list.
// Each "object" in a list is responsible for knowing how to compare itself.
int lookUp(void *data, List *list) {
   if (NULL == data)
      exit(-1);  // C does not have exceptions

   if (NULL == list->head)    // empty list
      return 0;
   else if ((((struct Node*)(list->head))->object->equals)(((struct Node*)(list->head))->object->data, data))
      return 1;
   else
      return lookUp(data, tail(list)); // recursion
}


// This is a "generic" function that determines if two lists are equal.
// Each "object" in a list is responsible for knowing how to compare itself.
int equals(List *list1, List *list2) {
   if (list1->head == list2->head)
      return 1;
   else if (NULL == list1->head || NULL == list2->head) // one list is empty but the other isn't
      return 0;
   else if (! (((struct Node*)(list1->head))->object->equals)(((struct Node*)(list1->head))->object->data,
                                                              ((struct Node*)(list2->head))->object->data))
      return 0;
   else
      return equals(tail(list1), tail(list2)); // recursion
}


// This is a "generic" function that prints the data in a list. Each
// "object" in the list is responsible for knowing how to print itself.
void printList(List *list) {
   printf("[");
   if (NULL != list->head) {
      (((struct Node*)(list->head))->object->toString)(((struct Node*)(list->head))->object->data);
      list = tail(list);
   }
   while (NULL != list->head) {
      printf(", ");
      (((struct Node*)(list->head))->object->toString)(((struct Node*)(list->head))->object->data);
      list = tail(list);
   }
   printf("]\n");
}
