/*
   This file implements a generic list with polymorphic functions.

   This file shows how to use "pointers to void", along with
   "function pointers", to implement a "generic", "polymorphic"
   list in C.

   Every list can hold any type of data, so if you
   want to, you can mix up ints, doubles and strings
   in a single list. This version of the list is able
   to print out any list, as long as each "object" in
   the list knows how to print itself. Each object in
   the list must have a pointer to a "toString" function
   that knows how to print the data in that object.

   This example gives you a very rough idea of how
   polymorphic methods are implemented by an
   Object-Oriented Programming language (they use
   a lot of function pointers).
*/
#include <stdio.h>
#include <stdlib.h> // exit()
#include "List.h"


List* newList() {  // construct an empty List
   List *list = (List*)malloc(sizeof(List));
   list->head = NULL;
   list->tail = NULL;
   return list;
}


void freeList(List *list) {
   // deallocate the list
   while (NULL != list) {
      List *temp = list->tail;
      if (NULL != list->head)
         free(list->head);
      free(list);
      list = temp;
   }
   free(list);
}


int empty(List *list) {
   return NULL == list->head;
}


struct Object* head(List *list) {
   if (NULL == list->head)
      exit(-1);  // C does not have exceptions

   return list->head;
}


List* tail(List *list) {
   if (NULL == list->head)
      exit(-1);  // C does not have exceptions

   return list->tail;
}


List* cons(struct Object *object, List *list) {
   List *list2 = newList(); // a new empty list
   list2->head = object;
   list2->tail = list;
   return list2;
}


int length(List *list) {
   if (empty(list))
      return 0;
   else
      return 1 + length(tail(list));
}


// This is a "generic" function that determines if a value is in a list.
// Each "object" in a list is responsible for knowing how to compare itself.
int lookUp(void *data, List *list)
{
   if (NULL == data)
      exit(-1);  // C does not have exceptions

   if (empty(list))
      return 0;
   else if ((head(list)->equals)(head(list)->data, data))
      return 1;
   else
      return lookUp(data, tail(list));
}


// This is a "generic" function that prints the data in a list. Each
// "object" in the list is responsible for knowing how to print itself.
void printList(List *list)
{
   printf("[");
   if (! empty(list)) {
      (head(list)->toString)(head(list)->data);
      list = tail(list);
   }
   while (! empty(list)) {
      printf(", ");
      (head(list)->toString)(head(list)->data);
      list = tail(list);
   }
   printf("]\n");
}


// This is a "generic" function that determines if two lists are equal.
// Each "object" in a list is responsible for knowing how to compare itself.
int equals(List *list1, List *list2)
{
   if (list1 == list2)
      return 1;
   else if (empty(list1) || empty(list2))
      return 0;
   else if (! (head(list1)->equals)(head(list1)->data, head(list2)->data))
      return 0;
   else
      return equals(tail(list1), tail(list2));
}
