/**
   These List objects CANNOT represent an empty list, so we
   need to use a null reference to represent an empty list.

   This List class is not very object-oriented.
   The List methods are all static methods.

   The List methods are static so that we can handle the
   case of the "empty list" (i.e., a null reference).
   If we define an empty List,
      List list1 = null;
   and if length() were an instance method in the List class,
   then
      list1.length();
   would not compute the length of list1 (which should be zero).
   Instead, we would get a null pointer exception. On the other
   hand, if length() is a static method in the List class, then
      List.length(list1);
   will work (see its definition below) and return the correct
   value, zero. So, in order to handle all the "empty list"
   cases, all the List methods need to be static methods in the
   List class. This would be a natural choice in the case of C
   (which is not object-oriented) and it is also the way some
   data structure textbooks define the List class.

   In this design we cannot implement the mutating List API,
      Object first()
      void rest()
      void insert(Object data)
   which are mutating versions of
      Object head()
      List tail()
      List cons(Object data)
*/
public class List
{
   private Object data;
   public List tail;

   List(Object data) { // construct a non-empty list of one item
      this(data, null);
   }


   List(Object data, List tail) { // construct a non-empty list of at least one item
      if (null == data)
         throw new IllegalArgumentException();

      this.data = data;
      this.tail = tail;
   }


   public static boolean empty(List list) {
      return null == list;
   }


   public static Object head(List list) {
      if (null == list)  // empty list
         throw new UnsupportedOperationException();

      return list.data;
   }


   public static List tail(List list) {
      if (null == list)  // empty list
         throw new UnsupportedOperationException();

      return list.tail;
   }


   public static List cons(Object data, List list) {
      return new List(data, list);
   }


   public static int length(List list) {
     if (null == list) // empty list
        return 0;
     else              // non-empty list
        return 1 + length(list.tail); // recursion
   }


   public static boolean lookUp(Object data, List list) {
      if (null == data)
         throw new IllegalArgumentException();

      if (null == list)                  // empty list
         return false;
      else if (data.equals(list.data))   // non-empty list
         return true;
      else
         return lookUp(data, list.tail); // recursion
   }


   public static boolean equals(List list1, List list2) {
      if (list1 == list2)
         return true;
      else if (null == list1 || null == list2) // one list is empty but the other isn't
         return false;
      else if (! list1.data.equals(list2.data))
         return false;
      else
         return equals(list1.tail, list2.tail); // recursion
   }


   // Convert list to Haskell's recursive notation for a list.
   public static String toStringRec(List list) {
      if (null == list) { // empty list
         return "[]";
      }
      else { // non-empty list
         return list.data + " : " + toStringRec(list.tail); // recursion
      }
   }


   public static String toString(List list) {
      String result = "[";
      if (null != list) {
         result += list.data;
         list = list.tail;
      }
      while (null != list) {
         result += ", " + list.data;
         list = list.tail;
      }
      result += "]";
      return result;
   }
}
