/**
   These List objects CAN represent an empty list (this.head == null).
   But we need a (private) Node class to implement this List.

   This List class is more object-oriented than the previous
   design. All the List methods are now instance methods.

   This List class DOES NOT have a recursive structure.
   In particular, notice that we can't write recursive
   length() or lookUp() methods! (It is the private Node
   class that turns out to be a recursive data type.)

   Notice that in this design we can implement the mutating List API,
      Object first()
      void rest()
      void insert(Object data)
   which are mutating versions of
      Object head()
      List tail()
      List cons(Object data)
*/
class List<T>
{
   private Node<T> head;

   List() { // construct an empty list
      head = null;
   }


   List(T data) { // construct a non-empty list of one item
      this(data, new List<T>());
   }


   List(T data, List<T> tail) { // construct a non-empty list of at least one item
      if (null == tail)
         throw new IllegalArgumentException();

      head = new Node<T>(data, tail.head);
   }


   public boolean empty() {
      return null == head;
   }


   public T head() {
      if (null == head) // empty list
         throw new UnsupportedOperationException();

      return head.data;
   }


   public List<T> tail() {
      if (null == head) // empty list
         throw new UnsupportedOperationException();

      List<T> result = new List<>();
      result.head = head.tail;
      return result;
   }


   public List<T> cons(T data) {
      List<T> result = new List<>();
      result.head = new Node<T>(data, head);
      return result;
   }


   public T first() { // mutating version of head()
      if (null == head)    // empty list
         throw new UnsupportedOperationException();

      T temp = head.data;
      head = head.tail;
      return temp;
   }


   public void rest() { // mutating version of tail()
      if (null == head) // empty list
         throw new UnsupportedOperationException();

      head = head.tail;
   }


   public void insert(T data) { // mutating version of cons()
      if (null == data)
         throw new IllegalArgumentException();

      head = new Node<T>(data, head);
   }


   // A non-recursive, iterative version of length()!
   public int length() {
      int result = 0;
      for (Node<T> cursor = head; null != cursor; cursor = cursor.tail) {
         result += 1;
      }
      return result;
   }

/*
   public int length() {
      if (null == head)
         return 0;
      else
         return 1 + length(????); // recursion will not work
   }
*/

   // A non-recursive, iterative version of lookUp()!
   public boolean lookUp(T data) {
      if (null == data)
         throw new IllegalArgumentException();

      for (Node<T> cursor = head; null != cursor; cursor = cursor.tail) {
         if (cursor.data.equals(data)) {
            return true;
         }
      }
      return false;
   }


   // A non-recursive, iterative version of equals()!
   public boolean equals(List<T> list) {
      if (this.head == list.head) return true;
      if (null == this.head || null == list.head) return false;
      Node<T> cursor1 = this.head;
      Node<T> cursor2 = list.head;
      while (null != cursor1 && null != cursor2) {
         if (! cursor1.data.equals(cursor2.data))
            return false;
         cursor1 = cursor1.tail;
         cursor2 = cursor2.tail;
      }
      if (null != cursor1 || null != cursor2) return false;
      return true;
   }


   // Convert this List to Haskell's recursive notation for a list.
   // But this has to be done iteratively, not recursively.
   public String toStringRec() {
      String result = "";
      for (Node<T> cursor = head; null != cursor; cursor = cursor.tail) {
         result += cursor.data + " : ";
      }
      result += "[]";
      return result;
   }


   public String toString() {
      String result = "[";
      if (null != head) {
         result += head.data;
         for (Node<T> cursor = head.tail; null != cursor; cursor = cursor.tail) {
            result += ", " + cursor.data;
         }
      }
      result += "]";
      return result;
   }
}



/**
   Notice that it is this Node class that is a recursive type.
*/
class Node<T>
{
   T data;
   Node<T> tail; // The tail is either empty (null) or the rest of the List.

   Node(T data, Node<T> tail) {
      if (null == data)
         throw new IllegalArgumentException();

      this.data = data;
      this.tail = tail;
   }
}
