/**
   These List objects CAN represent an empty list (this.head == null).
   But we need a (private) Node class to implement this List.

   This List class is more object-oriented than the previous
   design. All the List methods are now instance methods.

   This List class DOES NOT have a recursive structure.
   In particular, notice that we can't write recursive
   length() or lookUp() methods on the List data type!
   But the private Node class turns out to be a recursive
   data type and we can write in this List class recursive
   helper methods on the Node class.

   Notice that in this design we can implement the mutating List API,
      Object first()
      void rest()
      void insert(Object data)
   which are mutating versions of
      Object head()
      List tail()
      List cons(Object data)
*/
class List
{
   private Node head;

   List() { // construct an empty list
      head = null;
   }


   List(Object data) { // construct a non-empty list of one item
      this(data, new List());
   }


   List(Object data, List tail) { // construct a non-empty list of at least one item
      if (null == tail)
         throw new IllegalArgumentException();

      head = new Node(data, tail.head);
   }


   public boolean empty() {
      return null == head;
   }


   public Object head() {
      if (null == head) // empty list
         throw new UnsupportedOperationException();

      return head.data;
   }


   public List tail() {
      if (null == head) // empty list
         throw new UnsupportedOperationException();

      List result = new List();
      result.head = this.head.tail;
      return result;
   }


   public List cons(Object data) {
      List result = new List();
      result.head = new Node(data, this.head);
      return result;
   }


   public Object first() { // mutating version of head()
      if (null == head)    // empty list
         throw new UnsupportedOperationException();

      Object temp = head.data;
      head = head.tail;
      return temp;
   }


   public void rest() { // mutating version of tail()
      if (null == head) // empty list
         throw new UnsupportedOperationException();

      head = head.tail;
   }


   public void insert(Object data) { // mutating version of cons()
      head = new Node(data, head);
   }


   public int length() {
      return length(head); // recursive helper method
   }

   private int length(Node startNode) {
      if (null == startNode)  // empty list
         return 0;
      else                    // non-empty list
         return 1 + length(startNode.tail); // recursion
   }


   public boolean lookUp(Object data) {
      if (null == data)
         throw new IllegalArgumentException();

      return lookUp(data, head); // recursive helper method
   }

   private boolean lookUp(Object data, Node startNode) {
      if (null == startNode)                  // empty list
         return false;
      else if (startNode.data.equals(data))   // non-empty list
         return true;
      else
         return lookUp(data, startNode.tail); // recursion
   }


   public boolean equals(List list) {
      return equals(this.head, list.head); // recursive helper method
   }

   private boolean equals(Node list1, Node list2) {
      if (list1 == list2)
         return true;
      else if (null == list1 || null == list2)  // one list is empty
         return false;
      else if (! list1.data.equals(list2.data)) // neither list is empty
         return false;
      else
         return equals(list1.tail, list2.tail); // recursion
   }


   // Convert this List to Haskell's recursive notation for a list.
   public String toStringRec() {
      return toStringRec(head); // recursive helper method
   }

   private String toStringRec(Node startNode) {
      if (null == startNode) // empty list
         return "[]";
      else                   // non-empty list
         return startNode.data + " : " + toStringRec(startNode.tail);
   }


   public String toString() {
      String result = "[";
      if (null != head) {
         result += head.data;
         for (Node cursor = head.tail; null != cursor; cursor = cursor.tail) {
            result += ", " + cursor.data;
         }
      }
      result += "]";
      return result;
   }
}



/**
   Notice that it is this Node class that is a recursive type.
*/
class Node
{
   Object data;
   Node tail; // The tail is either empty (null) or the rest of the List.

   Node(Object data, Node tail) {
      if (null == data)
         throw new IllegalArgumentException();

      this.data = data;
      this.tail = tail;
   }
}
