/**
   These List objects CAN represent an empty list (this.head == null).
   But we need a (private) Node class to implement this List.

   This List class is more object-oriented than the previous
   design. All the List methods are now instance methods.

   This List class DOES NOT have a recursive structure.
   In particular, notice that we can't write recursive
   length() or lookUp() methods on the List data type!
   But the private Node class turns out to be a recursive
   data type and we can write recursive methods in the
   Node class and call them from within the List class.

   Notice that in this design we can implement the mutating List API,
      Object first()
      void rest()
      void insert(Object data)
   which are mutating versions of
      Object head()
      List tail()
      List cons(Object data)
*/
class List
{
   private Node head;

   List() { // construct an empty list
      head = null;
   }


   List(Object data) { // construct a non-empty list of one item
      this(data, new List());
   }


   List(Object data, List tail) { // construct a non-empty list of at least one item
      if (null == tail)
         throw new IllegalArgumentException();

      head = new Node(data, tail.head);
   }


   public boolean empty() {
      return null == head;
   }


   public Object head() {
      if (null == head) // empty list
         throw new UnsupportedOperationException();

      return head.data;
   }


   public List tail() {
      if (null == head) // empty list
         throw new UnsupportedOperationException();

      List result = new List();
      result.head = this.head.tail;
      return result;
   }


   public List cons(Object data) {
      List result = new List();
      result.head = new Node(data, this.head);
      return result;
   }


   public Object first() { // mutating version of head()
      if (null == head)    // empty list
         throw new UnsupportedOperationException();

      Object temp = head.data;
      head = head.tail;
      return temp;
   }


   public void rest() { // mutating version of tail()
      if (null == head) // empty list
         throw new UnsupportedOperationException();

      head = head.tail;
   }


   public void insert(Object data) { // mutating version of cons()
      head = new Node(data, head);
   }


   public int length() {
      if (null == head)  // empty list
         return 0;
      else               // non-empty list
         return 1 + head.length();
   }


   public boolean lookUp(Object data) {
      if (null == data)
         throw new IllegalArgumentException();

      if (null == head)   // empty list
         return false;
      else                // non-empty list
         return head.lookUp(data);
   }


   public boolean equals(List list) {
      if (this.head == list.head) return true;
      if (null == this.head || null == list.head) return false;
      return this.head.equals(list.head);
   }


   // Convert this List to Haskell's recursive notation for a list.
   public String toStringRec() {
      if (null == head) // empty list
         return "[]";
      else              // non-empty list
         return head.toStringRec();
   }


   public String toString() {
      String result = "[";
      if (null != head) {
         result += head.data;
         for (Node cursor = head.tail; null != cursor; cursor = cursor.tail) {
            result += ", " + cursor.data;
         }
      }
      result += "]";
      return result;
   }
}



/**
   Notice that it is this Node class that is a recursive type.

   It contains several recursive helper methods that are called
   by the equivalent List methods.
*/
class Node
{
   Object data;
   Node tail; // The tail is either empty (null) or the rest of the List.

   Node(Object data, Node tail) {
      if (null == data)
         throw new IllegalArgumentException();

      this.data = data;
      this.tail = tail;
   }

   public int length() {
      if (null == this.tail)
         return 1;
      else
         return 1 + this.tail.length(); // recursion
   }

   public boolean lookUp(Object data) {
      if (this.data.equals(data))  // non-empty list
         return true;
      else if (null == this.tail)
         return false;
      else
         return this.tail.lookUp(data); // recursion
   }

   public boolean equals(Node list2) {  // assume that list2 != null
      if (! this.data.equals(list2.data))
         return false;
      else if (this.tail == list2.tail)
         return true;
      else if (null == this.tail || null == list2.tail)
         return false;
      else
         return this.tail.equals(list2.tail); // recursion
   }

   // Convert this Node to Haskell's recursive notation for a list.
   public String toStringRec() {
      String result = this.data + " : ";
      if (null == this.tail) {
         return result + "[]";
      }
      else {
         return result + this.tail.toStringRec(); // recursion
      }
   }
}
