/**
   This List is implementated using the idea of an
   "algebraic data type". This List is a "sum type".
   It is a kind of "tagged union" (or "discriminated
   union").

   A List object holds two fields, a reference to a
   data object and a referenece to another List object.
   If the data object reference is null, then the List
   object is interpreted as representing an empty list.
   This is the "tagged union" idea. The data reference
   is really serving two purposes. When it is non-null,
   it refers to a data item in the list. But if the data
   object reference is null, then it changes the meaning
   of the List object and instead of being an item in a
   list, the List object is actually the empty list.

   So the data field is a kind of "tag" that determines
   how the List object gets interpreted. But the data
   field is not just a tag, it is also the reference to
   a data item in the list.
*/
public class List<T>
{
   private T data;  // null means empty list
   private List<T> tail;


   List() { // construct an empty list
      data = null;
      tail = null;
   }


   List(T data) { // construct a non-empty list of one item
      this(data, new List<>());
   }


   List(T data, List<T> tail) { // construct a non-empty list of at least one item
      if (null == data || tail == null)
         throw new IllegalArgumentException();

      this.data = data;
      this.tail = tail;
   }


   public boolean empty() {
      return null == data;
   }


   public T head() {
      if (null == data)  // empty list
         throw new UnsupportedOperationException();

      return data;
   }


   public List<T> tail() {
      if (null == data)  // empty list
         throw new UnsupportedOperationException();

      return tail;
   }


   public List<T> cons(T data) {
      return new List<>(data, this);
   }


   public int length() {
     if (null == data)            // empty list
        return 0;
     else                         // non-empty list
        return 1 + tail.length(); // recursion
   }


   public boolean lookUp(T data) {
      if (null == data)
         throw new IllegalArgumentException();

      if (null == this.data)            // empty list
         return false;
      else if (this.data.equals(data))  // non-empty list
         return true;
      else
         return tail.lookUp(data);      // recursion
   }


   public boolean equals(List<T> list2) {
      if (this == list2)
         return true;
      else if (null == data && null == list2.data) // both empty
         return true;
      else if (null == data || null == list2.data) // one list is empty but the other isn't
         return false;
      else if (! data.equals(list2.data))
         return false;
      else
         return tail.equals(list2.tail); // recursion
   }


   // Convert list to Haskell's recursive notation for a list.
   public String toStringRec() {
      if (null == data) {                          // empty list
         return "[]";
      }
      else {                                       // non-empty list
         return data + " : " + tail.toStringRec(); // recursion
      }
   }


   public String toString() {
      String result = "[";
      List<T> cursor = this;
      if (null != cursor.data) {
         result += cursor.data;
         cursor = cursor.tail;
      }
      while (null != cursor.data) {
         result += ", " + cursor.data;
         cursor = cursor.tail;
      }
      result += "]";
      return result;
   }
}
