/**
   This abstract generic class defines the interface for
   our List abstract data type (ADT).

   There will be two implementation of this interface,
   an "empty list" implementation, and a "non-empty list"
   implementation.

   In theory, we can characterize a List as follows,
      List<T> :: Empty
               | T : List<T>
   This is a recursive data type definition. In this
   implementation of the List ADT, the recursion in this
   definition is split over three files, the List<T> abstract
   class, the List_Nil<T> implementation, and the List_Node<T>
   implementation (which has an indirect recursive reference
   back to the List<T> class).

   In this design of List, as much of the implementation of the
   ADT as possible was done in this abstract base class. Notice
   that the abstract empty(), head(), and tail() methods are
   enough to write recursive implementations of most of the other
   methods from the ADT. That's because those three methods define
   the parts of this definition for the ADT.
      List<T> :: Empty         // empty()
               | T : List<T>   // head() : tail()
*/
public abstract class List<T>
{
   public abstract boolean empty();
   public abstract T head();
   public abstract List<T> tail();
   public abstract List<T> cons(T data);

   public int length() {
      if (empty())   // empty list
         return 0;
      else           // non-empty list
         return 1 + tail().length(); // is this recursion?
   }

   public boolean lookUp(T data) {
      if (null == data)
         throw new IllegalArgumentException();

      if (empty())   // empty list
         return false;
      else if (head().equals(data))
         return true;
      else
         return tail().lookUp(data); // is this recursion?
   }

   public boolean equals(List<T> list) {
      if (null == list)
         throw new IllegalArgumentException();

      if (this == list)
         return true;
      else if (empty() && list.empty()) // both lists are empty
         return true;
      else if (empty() || list.empty()) // one of the lists is empty
         return false;
      else if (! head().equals(list.head())) // neither list is empty
         return false;
      else
         return tail().equals(list.tail()); // is this recursion?
   }

   public abstract String toString();

   public String toStringRec() { // Haskell's list notation
      if (empty())
         return "[]";
      else
         return head() + " : " + tail().toStringRec(); // is this recursion?
   }
}

/*
   Notice a few things.
   1) This List ADT has a concrete notion of an empty list.
      It does not depend on nulls to denote empty. That is good.
   2) In fact, nulls cannot appear anywhere in these lists.
      That is enforced by the constructors, which is good.
   3) This List ADT is explicitly recursive, as it should be.
   4) But the recursive definition of the List ADT is split
      over three files. That's not good.
   5) The "recursive" methods in this file are not really
      recursive because of subtype polymorphism.
   6) The List_Nil<T> class is needlessly generic since
      it doesn't hold any data. This adds complexity
      where it's not needed.
   7) If we don't declare the two concrete implementations
      of this class as final, then we are open to some very
      strange bugs (see note (5)).

   The "algebraic data types" in Rust solve these problems.
*/
